/* Copyright (C) 2007 Christian Dywan <christian at twotoasts dot de>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. */

#ifndef __MIDORI_H__
#define __MIDORI_H__ 1

#include "config.h"

#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
#endif
#ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #include <signal.h>
#endif
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <libsexy/sexy.h>

#include "sokoke.h"

#ifdef MIDORI_WEBI
 #include <webi/webi.h>
#elif defined(MIDORI_WEBKIT)
 #include <WebCore/config.h>

 #include <Platform.h>
 #include <ChromeClientGdk.h>
 #include <ContextMenuClientGdk.h>
 #include <Document.h>
 #include <DragClient.h>
 #include <EditorClientGdk.h>
 #include <FrameGdk.h>
 #include <FrameLoader.h>
 #include <FrameLoaderClientGdk.h>
 #include <FrameView.h>
 #include <KURL.h>
 #include <Logging.h>
 #include <Page.h>
 #include <PlatformString.h>
 #include <ResourceHandleManager.h>
 
 #if SVG_SUPPORT
  #include "SVGNames.h"
  #include "XLinkNames.h"
  #include "SVGDocumentExtensions.h"
 #endif

 using namespace WebCore;
#else
 #include <webkit-gtk/webkit.h>
 #include <webi-external-content-delegate.h>
 #include <webi-auth.h>
 #include <webkit-gtk/webkit-view-webi-extensions.h>
#endif

// Our beautiful debug macros go here

static gchar DEBUG_MSG[256];
#ifndef __func__
 #define __func__ __FUNCTION__
#endif
#if MIDORI_DEBUG > 1
 #define DEBUG_FUNC g_snprintf(DEBUG_MSG, 255, "%s %s %d"\
  , __FILE__, __func__, __LINE__);g_print("%s\n", DEBUG_MSG);
#elif MIDORI_DEBUG > 0
 #define DEBUG_FUNC g_snprintf(DEBUG_MSG, 255, "%s %s %d"\
  , __FILE__, __func__, __LINE__);
#else
 #define DEBUG_FUNC ;
#endif

// For unimplemeted functions
#define UNIMPLEMETED g_print("Unimplemented: %s\n", __func__);

// Many themes need this hack for small toolbars to work
#define GTK_ICON_SIZE_SMALL_TOOLBAR GTK_ICON_SIZE_MENU

// -- Browser object

typedef struct _CBrowser
{
    // shared widgets
    GtkWidget* window;
    GtkWidget* menubar;
    GtkWidget* menu_closetab;
    GtkWidget* menu_tabs;
    GtkWidget* toolbar;
    GtkToolItem* toolbar_back;
    GtkToolItem* toolbar_forward;
    GtkToolItem* toolbar_refresh;
    GtkToolItem* toolbar_closetab;
    GtkToolItem* toolbar_tabtrash;
    GtkWidget* bookmarkbar;
    GtkWidget* icon_page;
    GtkWidget* toolbar_location;
    GtkWidget* icon_lock;
    GtkWidget* icon_search;
    GtkWidget* icon_search_add;
    GtkWidget* toolbar_search;
    GtkWidget* pages;
    GtkWidget* findbox;
    GtkWidget* findbox_text;
    GtkToolItem* findbox_case;
    GtkWidget* statusbar;
    GtkWidget* icon_link_script;
    GtkWidget* icon_link_window;
    GtkWidget* progress;
    GtkWidget* misc_feed;

    // tab specific widgets
    GtkWidget* page_icon;
    GtkWidget* page_name;
    #ifdef MIDORI_WEBKIT
    FrameGdk* gFrame;
    #endif
    GtkWidget* document;

    // tab specific values
    gboolean bMenubar;
    gboolean bToolbar;
    gboolean bLocation;
    gboolean bStatusbar;
    gboolean bResizable;
    int loadPercent;
    int bytesLoaded;
    int maxBytesLoaded;
    //UNDEFINED favicon;
    guint security;
    gchar* status_browser; // browser status messages
    gchar* status_document; // messages from a web resource
    gchar** resources; // list of loaded uris
} CBrowser;

// -- Resource object

typedef struct _CResource
{
    gchar* sUri;
    gchar* sTitle;
} CResource;

// -- User configuration object

typedef struct _CConfig
{
    gchar* sHomepage;
    gboolean bShowToolbar;
    gint iToolbarSize;
    gboolean bShowBookmarkbar;
    //gboolean bShowDownloadbar;
    gboolean bShowStatusbar;
    gint iTabPos;
    guint iTabMode; // 0: horizontal, 1: vertical
    guint iTabSize; // tab size in charcters
    guint iTabClose; // 0: no button, 1: button on every tab
    guint iPopupRule; // -> on_document_window_open
} CConfig;

// -- function declarations begin here

gboolean document_can_go_back(GtkWidget*, CBrowser*);
gboolean document_can_go_forward(GtkWidget*, CBrowser*);
G_CONST_RETURN gchar* document_get_uri(GtkWidget*, CBrowser*);
G_CONST_RETURN gchar* document_get_title(GtkWidget*, CBrowser*);
void document_set_zoom_level(GtkWidget*, gfloat, CBrowser*);
void on_document_popup(GtkWidget*, GdkEventButton*, CBrowser*);
static CBrowser* browser_new(CBrowser*);

#endif /* !__MIDORI_H__ */
